#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#

from enum import Enum

from pydantic import BaseModel, Field, RootModel

from cml_mcp.schemas.common import IPAddress, UUID4Type
from cml_mcp.schemas.node_definitions import LibvirtDomainDrivers
from cml_mcp.schemas.nodes import NodeLabel


UpperLibvirtDomainDrivers = Enum(
    "UpperLibvirtDomainDrivers",
    {member.name: member.name.upper() for member in LibvirtDomainDrivers},
)


class BaseLabDetails(BaseModel):
    node_id: UUID4Type
    compute_id: UUID4Type
    lab_id: UUID4Type
    label: NodeLabel


class ConsoleLabDetail(BaseLabDetails, extra="forbid"):
    driver: UpperLibvirtDomainDrivers
    line: int = Field(..., ge=0)


class VNCLabDetail(BaseLabDetails, extra="forbid"):
    compute_address: IPAddress


class ConsoleKeysResponse(RootModel[dict[UUID4Type, ConsoleLabDetail]]):
    pass


class VNCKeysResponse(RootModel[dict[UUID4Type, VNCLabDetail]]):
    pass



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#

from pydantic import BaseModel, Field


class ImportLabObject(BaseModel, extra="forbid"):
    id: str = Field(..., description="The lab ID of the imported lab.")
    warnings: list[str] | None = Field(..., description="Warnings, if any, as Markdown.")


class ImportLabResponse(ImportLabObject, extra="forbid"):
    """Successful lab import.
    The response has the lab ID and potential warnings as text / Markdown."""

    pass



# Copyright (c) 2025  Cisco Systems, Inc.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

import logging
import os
import tempfile
from typing import Any

import httpx
from fastmcp import Context, FastMCP
from mcp.shared.exceptions import McpError
from mcp.types import METHOD_NOT_FOUND
from virl2_client.models.cl_pyats import ClPyats

from cml_mcp.cml_client import CMLClient
from cml_mcp.schemas.annotations import (
    EllipseAnnotation,
    EllipseAnnotationResponse,
    LineAnnotation,
    LineAnnotationResponse,
    RectangleAnnotation,
    RectangleAnnotationResponse,
    TextAnnotation,
    TextAnnotationResponse,
)
from cml_mcp.schemas.common import UserName, UUID4Type
from cml_mcp.schemas.interfaces import InterfaceCreate, InterfaceResponse
from cml_mcp.schemas.labs import Lab, LabCreate, LabResponse, LabTitle

# from cml_mcp.schemas.licensing import LicensingStatus
from cml_mcp.schemas.links import Link, LinkCreate, LinkConditionConfiguration, ConditionResponse
from cml_mcp.schemas.node_definitions import NodeDefinition, SimplifiedNodeDefinitionResponse
from cml_mcp.schemas.nodes import Node, NodeConfigurationContent, NodeCreate, NodeLabel
from cml_mcp.schemas.system import SystemHealth, SystemInformation, SystemStats
from cml_mcp.schemas.topologies import Topology
from cml_mcp.settings import settings
from cml_mcp.types import Error

# Set up logging
logging.basicConfig(level=logging.DEBUG, format="%(asctime)s %(levelname)s %(threadName)s %(name)s: %(message)s")
logger = logging.getLogger("cml-mcp")


server_mcp = FastMCP(
    "Cisco Modeling Labs MCP Server",
    dependencies=["httpx", "fastmcp", "fastapi", "pydantic_strict_partial", "typer", "virl2_client", "pyats[full]"],
)
cml_client = CMLClient(host=str(settings.cml_url), username=settings.cml_username, password=settings.cml_password)


async def get_all_labs() -> list[UUID4Type]:
    """
    Get all labs from the CML server.

    Returns:
        list[UUID4Type]: A list of lab IDs.
    """
    labs = await cml_client.get("/labs", params={"show_all": True})
    return [UUID4Type(lab) for lab in labs]


@server_mcp.tool
async def get_cml_labs(user: UserName | None = None) -> list[Lab] | Error:
    """
    Get the list of labs for a specific user or all labs if the user is an admin.

    Args:
        user (UserName | None): The username to filter labs by. If None, defaults to the configured username.

    Returns:
        list[Lab] | Error: A list of Lab objects owned by the user, or an Error object if an error occurs.
    """
    # Clients like to pass "null" as a string vs. null as a None type.
    if not user or str(user) == "null":
        user = settings.cml_username  # Default to the configured username

    try:
        # If the requested user is not the configured user and is not an admin, deny access
        if str(user) != settings.cml_username and not await cml_client.is_admin():
            raise ValueError("User is not an admin and cannot view all labs.")
        ulabs = []
        # Get all labs from the CML server
        labs = await get_all_labs()
        for lab in labs:
            # For each lab, get its details
            lab_details = await cml_client.get(f"/labs/{lab}")
            # Only include labs owned by the specified user
            if lab_details.get("owner_username") == str(user):
                ulabs.append(Lab(**lab_details))
        return ulabs
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting CML labs: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_cml_information() -> SystemInformation | Error:
    """
    Get information about the CML server.

    Returns:
        SystemInformation: The system information.
        Error: An Error object if an exception occurs.
    """
    try:
        info = await cml_client.get("/system_information")
        return SystemInformation(**info)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting CML information: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_cml_status() -> SystemHealth | Error:
    """
    Get the status of the CML server.

    Returns:
        SystemHealth: The health/status of the CML server.
        Error: An Error object if an exception occurs.
    """
    try:
        status = await cml_client.get("/system_health")
        return SystemHealth(**status)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting CML status: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_cml_statistics() -> SystemStats | Error:
    """
    Get statistics about the CML server.

    Returns:
        SystemStats: The system statistics.
        Error: An Error object if an exception occurs.
    """
    try:
        stats = await cml_client.get("/system_stats")
        return SystemStats(**stats)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting CML statistics: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_cml_licensing_details() -> dict[str, Any] | Error:
    """
    Get licensing details from the CML server.

    Returns:
        dict[str, Any]: The licensing status as a dict corresponding to a LicensingStatus object.
        Error: An Error object if an exception occurs.
    """
    try:
        licensing_info = await cml_client.get("/licensing")
        # This is needed because some clients attempt to serialize the response
        # with Python classes for datetime rather than as pure JSON.  Cursor
        # is notably affected whereas Claude Desktop is not.
        return dict(licensing_info)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting CML licensing details: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_cml_node_definitions() -> list[SimplifiedNodeDefinitionResponse] | Error:
    """
    Get the list of node definitions from the CML server.

    Returns:
        list[SimplifiedNodeDefinitionResponse]: A list of SimplifiedNodeDefinitionResponse objects.
        Error: An Error object if an exception occurs.
    """
    try:
        node_definitions = await cml_client.get("/simplified_node_definitions")
        return [SimplifiedNodeDefinitionResponse(**nd) for nd in node_definitions]
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting CML node definitions: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


async def get_node_def_details(nid: UUID4Type) -> NodeDefinition:
    """
    Get detailed information about a specific node definition by its ID.

    Args:
        nid (UUID4Type): The node definition ID.

    Returns:
        NodeDefinition: The node definition details.
    """
    node_definition = await cml_client.get(f"/node_definitions/{nid}")
    return NodeDefinition(**node_definition)


@server_mcp.tool
async def get_node_definition_detail(nid: UUID4Type) -> NodeDefinition | Error:
    """
    Get detailed information about a specific node definition by its ID.

    Args:
        nid (UUID4Type): The node definition ID.

    Returns:
        NodeDefinition: The node definition details.
        Error: An Error object if an exception occurs.
    """
    try:
        return await get_node_def_details(nid)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting node definition detail for {nid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def create_empty_lab(lab: LabCreate | dict) -> LabResponse | Error:
    """
    Create an empty lab.

    Args:
        lab (LabCreate | dict): The empty lab definition as a LabCreate object or a dict
          corresponding to a LabCreate object.

    Returns:
        LabResponse: The created lab response.
        Error: An Error object if an exception occurs.
    """
    try:
        if isinstance(lab, dict):
            lab = LabCreate(**lab)
        resp = await cml_client.post("/labs", data=lab.model_dump(mode="json", exclude_defaults=True, exclude_none=True))
        return LabResponse(**resp)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error creating empty lab topology: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def create_full_lab_topology(topology: Topology | dict) -> UUID4Type | Error:
    """
    Create a new lab topology with lab details, nodes, links, node configurations, and other resources.

    Args:
        topology (Topology | dict): The topology definition as a Topology object or a dict
          corresponding to a Topology object.

    Returns:
        UUID4Type: The new lab topology ID if successful.
        Error: An Error object if an error occurs.
    """
    try:
        # Cursor passes a dict representing the schema of a Topology.
        # Handle this conversion.
        if isinstance(topology, dict):
            topology = Topology(**topology)
        resp = await cml_client.post("/import", data=topology.model_dump(mode="json", exclude_defaults=True))
        return UUID4Type(resp["id"])
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error creating lab topology: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def start_cml_lab(lid: UUID4Type) -> None | Error:
    """
    Start a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
        Error: An Error object if an error occurs.
    """
    try:
        await cml_client.put(f"/labs/{lid}/start")
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error starting CML lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


async def stop_lab(lid: UUID4Type) -> None:
    """
    Stop a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
    """
    await cml_client.put(f"/labs/{lid}/stop")


async def wipe_lab(lid: UUID4Type) -> None:
    """
    Wipe a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
    """
    await cml_client.put(f"/labs/{lid}/wipe")


@server_mcp.tool
async def stop_cml_lab(lid: UUID4Type) -> None | Error:
    """
    Stop a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
        Error: An Error object if an error occurs.
    """
    try:
        await stop_lab(lid)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error stopping CML lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def wipe_cml_lab(lid: UUID4Type, ctx: Context) -> None | Error:
    """
    Wipe a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
        Error: An Error object if an error occurs.
    """
    try:
        elicit_supported = True
        try:
            result = await ctx.elicit("Are you sure you want to wipe the lab?", response_type=None)
        except McpError as me:
            if me.error.code == METHOD_NOT_FOUND:
                elicit_supported = False
            else:
                raise me
        if not elicit_supported or result.action == "accept":
            await wipe_lab(lid)
        else:
            raise Exception("Wipe operation cancelled by user.")
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error wiping CML lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def delete_cml_lab(lid: UUID4Type, ctx: Context) -> None | Error:
    """
    Delete a CML lab by its ID. If the lab is running and/or not wiped, it will be stopped and wiped first.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        None: If successful.
        Error: An Error object if an error occurs.
    """
    try:

        elicit_supported = True
        try:
            result = await ctx.elicit("Are you sure you want to delete the lab?", response_type=None)
        except McpError as me:
            if me.error.code == METHOD_NOT_FOUND:
                elicit_supported = False
            else:
                raise me
        if not elicit_supported or result.action == "accept":
            await stop_lab(lid)  # Ensure the lab is stopped before deletion
            await wipe_lab(lid)  # Ensure the lab is wiped before deletion
            await cml_client.delete(f"/labs/{lid}")
        else:
            raise Exception("Delete operation cancelled by user.")
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error deleting CML lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


async def add_interface(lid: UUID4Type, intf: InterfaceCreate) -> InterfaceResponse:
    """
    Add an interface to a CML lab by its lab ID.

    Args:
        lid (UUID4Type): The lab ID.
        intf (InterfaceCreate): The interface definition as an InterfaceCreate object.

    Returns:
        InterfaceResponse: The added interface details if successful.
        Error: An Error object if an error occurs.
    """
    resp = await cml_client.post(f"/labs/{lid}/interfaces", data=intf.model_dump(mode="json", exclude_defaults=True))
    return InterfaceResponse(**resp)


@server_mcp.tool
async def add_node_to_cml_lab(lid: UUID4Type, node: NodeCreate | dict) -> UUID4Type | Error:
    """
    Add a node to a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.
        node (NodeCreate | dict): The node definition as a NodeCreate object or a dict
          corresponding to a NodeCreate object.

    Returns:
        UUID4Type: The ID of the added node if successful.
        Error: An Error object if an error occurs.
    """
    try:
        if isinstance(node, dict):
            node = NodeCreate(**node)
        resp = await cml_client.post(f"/labs/{lid}/nodes", data=node.model_dump(mode="json", exclude_defaults=True))
        # Create the default number of interfaces for the node def.  This is modeled after what happens in the UI.
        try:
            nd = await get_node_def_details(node.node_definition)
            defcount = nd.device.interfaces.default_count
            if not defcount:
                defcount = nd.device.interfaces.min_count
                if not defcount:
                    defcount = 1
            for _ in range(defcount):
                ic = {"node": resp["id"]}
                await add_interface(lid, InterfaceCreate(**ic))
        except Exception as ie:
            logger.error(f"Error adding interfaces for node {resp['id']} in lab {lid}: {str(ie)}", exc_info=True)
        return UUID4Type(resp["id"])
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error adding CML node to lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def add_annotation_to_cml_lab(
    lid: UUID4Type, annotation: TextAnnotation | RectangleAnnotation | EllipseAnnotation | LineAnnotation | dict
) -> TextAnnotation | RectangleAnnotation | EllipseAnnotation | LineAnnotation | Error:
    """Add an annotation to a CML lab by the lab ID.

    Args:
        lid (UUID4Type): The lab ID.
        annotation (TextAnnotation | RectangleAnnotation | EllipseAnnotation | LineAnnotation | dict): The annotation to add as
          a TextAnnotation, RectangleAnnotation, EllipseAnnotation, or LineAnnotation object, or a dict representing a
          TextAnnotation, RectangleAnnotation, EllipseAnnotation, or LineAnnotation.

    Returns:
        TextAnnotationResponse | RectangleAnnotationResponse | EllipseAnnotationResponse | LineAnnotationResponse | Error:
          The added annotation or an error.
    """
    try:
        if isinstance(annotation, dict):
            if "type" not in annotation:
                raise ValueError("Annotation dict must contain a 'type' field.")
            if annotation["type"] == "text":
                annotation = TextAnnotation(**annotation)
            elif annotation["type"] == "rectangle":
                annotation = RectangleAnnotation(**annotation)
            elif annotation["type"] == "ellipse":
                annotation = EllipseAnnotation(**annotation)
            elif annotation["type"] == "line":
                annotation = LineAnnotation(**annotation)
            else:
                raise ValueError(f"Unknown annotation type: {annotation['type']}")
        resp = await cml_client.post(f"/labs/{lid}/annotations", data=annotation.model_dump(mode="json", exclude_defaults=True))
        if resp["type"] == "text":
            return TextAnnotationResponse(**resp)
        elif resp["type"] == "rectangle":
            return RectangleAnnotationResponse(**resp)
        elif resp["type"] == "ellipse":
            return EllipseAnnotationResponse(**resp)
        elif resp["type"] == "line":
            return LineAnnotationResponse(**resp)
        else:
            raise ValueError(f"Unknown annotation type created: {resp['type']}")
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error adding annotation to lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def add_interface_to_node(lid: UUID4Type, intf: InterfaceCreate | dict) -> InterfaceResponse | Error:
    """
    Add an interface to a CML node by its lab ID.

    Args:
        lid (UUID4Type): The lab ID.
        intf (InterfaceCreate | dict): The interface definition as an InterfaceCreate object or a dict
          corresponding to an InterfaceCreate object.

    Returns:
        InterfaceResponse: The added interface details if successful.
        Error: An Error object if an error occurs.
    """
    try:
        if isinstance(intf, dict):
            intf = InterfaceCreate(**intf)
        return await add_interface(lid, intf)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error adding interface to node {intf.node} in lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_interfaces_for_node(lid: UUID4Type, nid: UUID4Type) -> list[InterfaceResponse] | Error:
    """
    Get interfaces for a specific node in a CML lab by its lab ID and node ID.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.

    Returns:
        list[NodeResponse]: A list of NodeResponse objects representing the node's interfaces.
        Error: An Error object if an error occurs.
    """
    try:
        resp = await cml_client.get(f"/labs/{lid}/nodes/{nid}/interfaces", params={"data": True, "operational": False})
        return [InterfaceResponse(**iface) for iface in resp]
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting interfaces for node {nid} in lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def connect_two_nodes(lid: UUID4Type, link_info: LinkCreate | dict) -> UUID4Type | Error:
    """
    Create a link between two nodes in a CML lab by their interface IDs.

    Args:
        link_info (LinkCreate | dict): Details for the link to create as a LinkCreate object or a dict
          corresponding to a LinkCreate object.

    Returns:
        UUID4Type: The ID of the link created if successful.
        Error: An Error object if an error occurs.
    """
    try:
        if isinstance(link_info, dict):
            link_info = LinkCreate(**link_info)
        resp = await cml_client.post(f"/labs/{lid}/links", data=link_info.model_dump(mode="json"))
        return UUID4Type(resp["id"])
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error creating link for {link_info}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_all_links_for_lab(lid: UUID4Type) -> list[Link] | Error:
    """
    Get all links for a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        list[Link]: List of Link objects.
        Error: An Error object if an error occurs.
    """
    try:
        resp = await cml_client.get(f"/labs/{lid}/links", params={"data": True})
        return [Link(**link) for link in resp]
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting links for lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def apply_conditioning_to_link(
    lid: UUID4Type, link_id: UUID4Type, condition: LinkConditionConfiguration | dict
) -> ConditionResponse | Error:
    """
    Condition a link in a CML lab by its lab ID and link ID.

    Args:
        lid (UUID4Type): The lab ID.
        link_id (UUID4Type): The link ID.
        condition (LinkConditionConfiguration): The link conditioning config to apply as a LinkConditionConfiguration object or a dict
          corresponding to a LinkConditionConfiguration object.

    Returns:
        ConditionResponse: If successful.
        Error: An Error object if an error occurs.
    """
    try:
        if isinstance(condition, dict):
            condition = LinkConditionConfiguration(**condition)
        resp = await cml_client.patch(f"/labs/{lid}/links/{link_id}/condition", data=condition.model_dump(mode="json", exclude_none=True))
        return ConditionResponse(**resp)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error conditioning link {link_id} in lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def configure_cml_node(lid: UUID4Type, nid: UUID4Type, config: NodeConfigurationContent) -> UUID4Type | Error:
    """
    Configure a node in a CML lab by its lab ID and node ID. The node must be in a BOOTED (i.e., wiped) state.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.
        config (NodeConfigurationContent): The configuration content for the node.

    Returns:
        UUID4Type: The ID of the node configured if successful.
        Error: An Error object if an error occurs.
    """
    payload = {"configuration": str(config)}
    try:
        resp = await cml_client.patch(f"/labs/{lid}/nodes/{nid}", data=payload)
        return UUID4Type(resp)
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error configuring CML node {nid} in lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_nodes_for_cml_lab(lid: UUID4Type) -> list[Node] | Error:
    """
    Get a list of nodes for a CML lab by its ID.

    Args:
        lid (UUID4Type): The lab ID.

    Returns:
        list[Node]: List of node objects (without their configurations).
        Error: An Error object if an error occurs.
    """
    try:
        resp = await cml_client.get(f"/labs/{lid}/nodes", params={"data": True, "operational": True, "exclude_configurations": True})
        rnodes = []
        for node in list(resp):
            # XXX: Fixup known issues with bad data coming from
            # certain node types.
            if node["operational"].get("vnc_key") == "":
                node["operational"]["vnc_key"] = None
            if node["operational"].get("image_definition") == "":
                node["operational"]["image_definition"] = None
            if node["operational"].get("serial_consoles") is None:
                node["operational"]["serial_consoles"] = []
            rnodes.append(Node(**node))
        return rnodes
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting nodes for CML lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def get_cml_lab_by_title(title: LabTitle) -> Lab | Error:
    """
    Get a CML lab by its title.

    Args:
        title (LabTitle): The lab title.

    Returns:
        Lab: The lab object if found.
        Error: An Error object if an error occurs or lab is not found.
    """
    try:
        labs = await get_all_labs()
        for lid in labs:
            lab = await cml_client.get(f"/labs/{lid}")
            if lab["lab_title"] == str(title):
                return Lab(**lab)
        raise ValueError(f"Lab with title '{title}' not found.")
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error getting CML lab by title {title}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def stop_cml_node(lid: UUID4Type, nid: UUID4Type) -> None | Error:
    """
    Stop a node in a CML lab by its lab ID and node ID.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.

    Returns:
        None: If successful.
        Error: An Error object if an error occurs.
    """
    try:
        await cml_client.put(f"/labs/{lid}/nodes/{nid}/state/stop")
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error stopping CML node {nid} in lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def start_cml_node(lid: UUID4Type, nid: UUID4Type) -> None | Error:
    """
    Start a node in a CML lab by its lab ID and node ID.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.

    Returns:
        None: If successful.
        Error: An Error object if an error occurs.
    """
    try:
        await cml_client.put(f"/labs/{lid}/nodes/{nid}/state/start")
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error starting CML node {nid} in lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
async def wipe_cml_node(lid: UUID4Type, nid: UUID4Type, ctx: Context) -> None | Error:
    """
    Wipe a node in a CML lab by its lab ID and node ID. Node must be stopped first.

    Args:
        lid (UUID4Type): The lab ID.
        nid (UUID4Type): The node ID.

    Returns:
        None: If successful.
        dict[str, str]: An error dictionary if an error occurs.
    """
    try:
        elicit_supported = True
        try:
            result = await ctx.elicit("Are you sure you want to wipe the node?", response_type=None)
        except McpError as me:
            if me.error.code == METHOD_NOT_FOUND:
                elicit_supported = False
            else:
                raise me
        if not elicit_supported or result.action == "accept":
            await cml_client.put(f"/labs/{lid}/nodes/{nid}/wipe_disks")
        else:
            raise Exception("Wipe operation cancelled by user.")
    except httpx.HTTPStatusError as e:
        return Error(**{"error": f"HTTP error {e.response.status_code}: {e.response.text}"})
    except Exception as e:
        logger.error(f"Error wiping CML node {nid} in lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})


@server_mcp.tool
def send_cli_command(lid: UUID4Type, label: NodeLabel, commands: str, config_command: bool = False) -> str | Error:
    """
    Send CLI command(s) to a node in a CML lab by its lab ID and node label.

    Args:
        lid (UUID4Type): The lab ID.
        label (NodeLabel): The label of the node to send the command to.
        commands (str): The CLI command(s) to send.  Multiple commands can be specified separated by newlines.
        config_command (bool, optional): If True, send as a configuration command(s).
         This will automatically put the device in config mode. Defaults to False, meaning command(s) must be in exec mode.

    Returns:
        str | Error: The command output if successful, or an Error object if an error occurs.
    """
    cwd = os.getcwd()  # Save the current working directory
    try:
        os.chdir(tempfile.gettempdir())  # Change to a writable directory (required by pyATS/ClPyats)
        lab = cml_client.vclient.join_existing_lab(str(lid))  # Join the existing lab using the provided lab ID
        pylab = ClPyats(lab)  # Create a ClPyats object for interacting with the lab
        pylab.sync_testbed(settings.cml_username, settings.cml_password)  # Sync the testbed with CML credentials
        if config_command:
            # Send the command as a configuration command
            return pylab.run_config_command(str(label), commands)
        # Send the command as an exec/operational command
        return pylab.run_command(str(label), commands)
    except Exception as e:
        logger.error(f"Error sending CLI command '{commands}' to node {label} in lab {lid}: {str(e)}", exc_info=True)
        return Error(**{"error": str(e)})
    finally:
        os.chdir(cwd)  # Restore the original working directory



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from __future__ import annotations

from enum import Enum
from typing import NewType

AnnotationId = NewType("AnnotationId", str)
BridgeId = NewType("BridgeId", str)
ComputeId = NewType("ComputeId", str)
ConsoleKey = NewType("ConsoleKey", str)
DomainId = NewType("DomainId", str)
ElementId = NewType("ElementId", str)
ExternalConnectorId = NewType("ExternalConnectorId", str)
GroupId = NewType("GroupId", str)
InterfaceId = NewType("InterfaceId", ElementId)
LabId = NewType("LabId", str)
RepoId = NewType("RepoId", str)
LinkCaptureKey = NewType("LinkCaptureKey", str)
LinkId = NewType("LinkId", ElementId)
MacAddress = NewType("MacAddress", str)
NodeId = NewType("NodeId", ElementId)
NoticeId = NewType("NoticeId", str)
ResourcePoolId = NewType("ResourcePoolId", str)
SmartAnnotationId = NewType("SmartAnnotationId", str)
Tag = NewType("Tag", str)
UserId = NewType("UserId", str)
UserName = NewType("UserName", str)
VNCKey = NewType("VNCKey", str)

ServiceKey = ConsoleKey | LinkCaptureKey | VNCKey


class Service(Enum):
    CONSOLE = "Console"
    PCAP = "PCAP"
    VNC = "VNC"



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from __future__ import annotations

import time
from datetime import datetime
from enum import Enum
from typing import NamedTuple

import attr

from .type_hints import (
    ComputeId,
    ExternalConnectorId,
    LabId,
    LinkCaptureKey,
    LinkId,
    NodeId,
    NoticeId,
    UserId,
)


class BootEventType(Enum):
    MONITOR = 1
    BOOTED = 2


@attr.s(kw_only=True)
class BaseEvent:
    timestamp: int = attr.ib(factory=lambda: int(time.time()))
    isotimestamp: str = attr.ib(init=False)

    def __attrs_post_init__(self):
        self.isotimestamp = datetime.fromtimestamp(self.timestamp).isoformat()


@attr.s(kw_only=True)
class BootProgressDiagnosticEvent(BaseEvent):
    lab_id: LabId = attr.ib()
    node_id: NodeId = attr.ib()
    event: BootEventType = attr.ib()

    def as_dict(self):
        return {
            "lab_id": self.lab_id,
            "node_id": self.node_id,
            "event": self.event.name,
            "timestamp": self.isotimestamp,
        }


class LabEventType(Enum):
    ADD = 1
    REMOVE = 2
    CHANGE = 3
    STATE_CHANGED = 4


class LabEventElementType(Enum):
    LAB = 1
    NODE = 2
    LINK = 3
    INTERFACE = 4
    ANNOTATION = 5
    CONNECTOR_MAPPING = 6
    SMART_ANNOTATION = 7


class LinkConditionEvent(NamedTuple):
    lab_id: LabId
    link_id: LinkId
    data: dict[str, int] = {}


class LldEventTypes(Enum):
    REGISTERED = 0
    CONNECTED = 1
    SYNCED = 2
    DISCONNECTED = 3
    READINESS = 4
    SYNC_FAILED = 5
    REMOVED = 6
    STATE_CHANGED = 7


class LldEvent(NamedTuple):
    id: ComputeId
    event_type: LldEventTypes
    data: dict = {}


class NodeEventTypes(Enum):
    DEFINED = 0
    UNDEFINED = 1
    STARTED = 2
    STOPPED = 5


class LldNodeEvent(NamedTuple):
    id: ComputeId
    node_id: NodeId
    event_type: NodeEventTypes


class NoticeEventTypes(Enum):
    ADD = 1
    REMOVE = 2
    CHANGE = 3
    ACKNOWLEDGEMENT = 4
    MAINTENANCE = 5


class NoticeEvent(NamedTuple):
    id: NoticeId
    event_type: NoticeEventTypes


class ExternalConnectorEventTypes(Enum):
    ADD = 1
    REMOVE = 2
    CHANGE = 3
    SYNC = 4


class ExternalConnectorEvent(NamedTuple):
    id: ExternalConnectorId | None
    event_type: ExternalConnectorEventTypes


class PcapEventTypes(str, Enum):
    START = 1
    STOP = 2


class PcapEvent(NamedTuple):
    status: PcapEventTypes
    lab_id: LabId
    link_id: LinkId
    link_capture_key: LinkCaptureKey


class StopConsoleEvent(NamedTuple):
    lab_id: LabId
    console_uuid: str


class SystemEventTypes(Enum):
    FEEDBACK_SUBMITTED = 1
    FEEDBACK_FAILURE = 2


class SystemEvent(NamedTuple):
    event_type: SystemEventTypes
    message: str
    data: dict


class LabUsersChangedEvent(NamedTuple):
    lab_id: LabId
    users: list[UserId]



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
import re
from typing import Annotated

from fastapi import Body
from pydantic import BaseModel, Field

from cml_mcp.schemas.common import Tag, TagArray

ExternalConnectorDeviceName = Annotated[
    str,
    Field(
        pattern=re.compile(r"^(bridge|local|virbr|vlan)\d{1,4}(?![\n\r])$"),
        description="A Linux bridge name usable for external connectivity.",
    ),
]

NullableExternalConnectorDeviceName = Annotated[
    ExternalConnectorDeviceName | None,
    Field(description="A nullable Linux bridge name usable for external connectivity."),
]

ExternalConnectorTag = Annotated[
    Tag, Field(description="The key configured in external connector nodes.")
]


class ExternalConnectorMappingBase(BaseModel, extra="allow"):
    """Base for mapping of external connector lab node configs to available bridges."""

    key: ExternalConnectorTag = Field(default=None)
    device_name: NullableExternalConnectorDeviceName = Field(default=None)


class ExternalConnectorMapping(ExternalConnectorMappingBase, extra="forbid"):
    """Mapping of external connector lab node configs to available bridges."""

    key: ExternalConnectorTag = Field(...)
    device_name: NullableExternalConnectorDeviceName = Field(...)
    label: str = Field(
        default=None,
        description="Unique label for the external connector.",
        max_length=128,
    )
    tags: TagArray = Field(
        default=None, description="Tags denoting purpose of the external connector."
    )
    allowed: bool = Field(
        default=None,
        description="""
      If true, the calling user is allowed to start external connector nodes
      which are configured to use this external connector mapping.
      Users may be limited by the resource pool settings imposed on them.
    """,
    )


class ExternalConnectorMappingUpdate(ExternalConnectorMappingBase, extra="forbid"):
    """Update an external connector mapping."""

    key: ExternalConnectorTag = Field(...)
    device_name: NullableExternalConnectorDeviceName = Field(...)


ExternalConnectorMappingBody = Annotated[
    list[ExternalConnectorMappingUpdate],
    Body(description="Partial list of external connector key-device_name mappings."),
]

ExternalConnectorMappingResponse = Annotated[
    list[ExternalConnectorMapping],
    Field(description="List of external connector key-device_name mappings."),
]



# Copyright (c) 2025  Cisco Systems, Inc.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

from cml_mcp.server import server_mcp


def main():
    server_mcp.run()


if __name__ == "__main__":
    main()



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from datetime import datetime
from typing import Annotated, Literal

from fastapi import Body
from pydantic import BaseModel, Field

from cml_mcp.schemas.simple_core.common.events import LabEventElementType, LabEventType
from cml_mcp.schemas.simple_core.common.states import InterfaceState, LinkState, NodeState
from cml_mcp.schemas.annotations import AnnotationResponse
from cml_mcp.schemas.common import (
    BaseDBModel,
    EffectivePermissions,
    GenericDescription,
    OldPermission,
    Permissions,
    State,
    UserFullName,
    UserName,
    UUID4Type,
)
from cml_mcp.schemas.links import Link
from cml_mcp.schemas.nodes import Node
from cml_mcp.schemas.smart_annotations import SmartAnnotationBase

LabTitle = Annotated[
    str,
    Field(
        min_length=1,
        max_length=64,
        description="Title of the Lab.",
        examples=["Lab at Mon 17:27 PM"],
    ),
]

LabNotes = Annotated[
    str,
    Field(
        max_length=32768,
        description="Additional, textual free-form Lab notes.",
        examples=["Find why this topology does not perform as expected!"],
    ),
]

LabDescription = Annotated[
    GenericDescription,
    Field(
        description="Additional, textual free-form detail of the lab.",
        examples=["CCNA study lab"],
    ),
]


class LabGroupBase(BaseModel, extra="allow"):
    id: UUID4Type = Field(..., description="ID of the lab group.")
    permission: OldPermission = Field(
        ..., description="Permission level for the lab group."
    )


class GroupLab(LabGroupBase, extra="forbid"):
    pass


class LabGroup(LabGroupBase, extra="forbid"):
    name: str = Field(default=None, description="Name of the lab group.")


class LabGroupAssociation(BaseModel, extra="forbid"):
    id: UUID4Type = Field(..., description="ID of the group.")
    permissions: Permissions = Field(
        ..., description="Permissions for the specified group and lab."
    )


class LabUserAssociation(BaseModel, extra="forbid"):
    id: UUID4Type = Field(..., description="ID of the user.")
    permissions: Permissions = Field(
        ..., description="Permissions for the specified user and lab."
    )


class LabAssociations(BaseModel, extra="forbid"):
    groups: list[LabGroupAssociation] = Field(
        default=None, description="Array of group associations."
    )
    users: list[LabUserAssociation] = Field(
        default=None, description="Array of user associations."
    )


LabOwner = Annotated[UUID4Type, Field(description="ID of the lab owner.")]


class LabCreate(BaseModel, extra="forbid"):
    """Lab metadata."""

    title: LabTitle = Field(default=None)
    owner: LabOwner = Field(default=None)
    description: LabDescription = Field(default=None)
    notes: LabNotes = Field(default=None)
    groups: list[LabGroup] = Field(
        default=None,
        deprecated=True,
        description="Array of LabGroup objects - mapping from group ID to permissions.",
    )
    associations: LabAssociations = Field(
        default=None,
        description="Object of lab/group and lab/user associations.",
    )


class Lab(BaseDBModel):
    """Metadata about the state of the lab itself."""

    lab_description: LabDescription = Field(default=None)
    lab_notes: LabNotes = Field(default=None)
    lab_title: LabTitle = Field(...)
    owner: LabOwner = Field(default=None)
    owner_username: UserName = Field(default=None, description="The owner username.")
    owner_fullname: UserFullName = Field(
        default=None, description="The owner full name."
    )
    state: State = Field(..., description="The overall state of the lab.")
    node_count: int = Field(
        default=None, description="Number of nodes (or devices) in the lab.", ge=0
    )
    link_count: int = Field(
        default=None,
        description="Number of connections between nodes in the lab.",
        ge=0,
    )
    groups: list[LabGroup] = Field(
        default=None,
        deprecated=True,
        description="Array of LabGroup objects - mapping from group id to permissions.",
    )
    effective_permissions: EffectivePermissions = Field(...)


LabGroupsBody = Annotated[
    list[LabGroup],
    Body(
        description="This request body is a JSON object that describes the group permissions for a lab."
    ),
]

LabCreateBody = Annotated[LabCreate, Body(description="The lab's data.")]


class LabResponse(Lab, extra="forbid"):
    """The response body is a JSON lab object."""

    pass


LabNodeStateNames = list(NodeState.__members__.keys())
LabLinkStateNames = list(LinkState.__members__.keys())
LabInterfaceStateNames = list(InterfaceState.__members__.keys())


class LabElementStateResponse(BaseModel):
    """
    JSON object with the state of all
    nodes, interfaces, and links in the lab.
    """

    nodes: dict[UUID4Type, Literal[*LabNodeStateNames]] = Field(...)
    links: dict[UUID4Type, Literal[*LabLinkStateNames]] = Field(...)
    interfaces: dict[UUID4Type, Literal[*LabInterfaceStateNames]] = Field(...)


class LabSimulationStatsResponse(BaseModel):
    nodes: list[Node] = Field(default=None)
    links: list[Link] = Field(default=None)


LabEventNames = list(LabEventType.__members__.keys())
LabEventElementTypeNames = list(LabEventElementType.__members__.keys())


class LabEventResponse(BaseModel):
    """Lab Event Info"""

    lab_id: UUID4Type = Field(...)
    event: Literal[*LabEventNames] = Field(...)
    element_type: Literal[*LabEventElementTypeNames] = Field(...)
    element_id: UUID4Type = Field(...)
    data: dict = Field(...)
    previous: dict = Field(...)
    timestamp: datetime = Field(...)


class SimplifiedLabTopology(BaseModel):
    nodes: list[Node] = Field(...)
    links: list[Link] = Field(...)
    annotations: list[AnnotationResponse] = Field(default=None)
    smart_annotations: list[SmartAnnotationBase] = Field(default=None)


class LabInfoResponse(BaseDBModel, extra="forbid"):
    state: State = Field(..., description="The overall state of the lab.")
    lab_title: LabTitle = Field(...)
    lab_description: LabDescription = Field(...)
    lab_notes: LabNotes = Field(...)
    owner: LabOwner = Field(...)
    owner_username: str = Field(...)
    owner_fullname: str = Field(...)
    node_count: int = Field(..., ge=0, description="Number of nodes in the lab.")
    link_count: int = Field(
        ..., ge=0, description="Number of connections between nodes in the lab."
    )
    groups: list[LabGroup] = Field(...)
    effective_permissions: EffectivePermissions = Field(...)
    topology: SimplifiedLabTopology = Field(
        default=None, description="Lab topology data"
    )



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from typing import Annotated

from fastapi import Body
from pydantic import BaseModel, Field

from cml_mcp.schemas.common import Label, State, UUID4Type

LinkLabel = Annotated[Label, Field(..., description="A link label.")]


class LinkCreate(BaseModel, extra="forbid"):
    src_int: UUID4Type = Field(...)
    dst_int: UUID4Type = Field(...)


class LinkConnectionInfo(BaseModel, extra="forbid"):
    interface_a: UUID4Type = Field(default=None, description="ID of the interface A.")
    interface_b: UUID4Type = Field(default=None, description="ID of the interface B.")
    node_a: UUID4Type = Field(default=None, description="ID of the node A.")
    node_b: UUID4Type = Field(default=None, description="ID of the node B.")


class Link(LinkConnectionInfo, extra="forbid"):
    """Link object."""

    id: UUID4Type = Field(default=None, description="ID of the link.")
    label: LinkLabel = Field(default=None)
    lab_id: UUID4Type = Field(default=None, description="ID of the lab.")
    link_capture_key: str = Field(default=None, max_length=64, description="The link capture key.")
    state: State = Field(default=None, description="The status of the link in the lab.")


class LinkCondition(BaseModel):
    bandwidth: int | None = Field(default=None, description="Bandwidth of the link in kbps.", ge=0, le=10000000)
    latency: int | None = Field(default=None, description="Delay of the link in ms.", ge=0, le=10000)
    delay_corr: float | int | None = Field(default=None, description="Loss correlation in percent.", ge=0, le=100)
    limit: int | None = Field(default=None, description="Limit in ms.", ge=0, le=10000)
    loss: float | int | None = Field(default=None, description="Loss of the link in percent.", ge=0, le=100)
    loss_corr: float | int | None = Field(default=None, description="Loss correlation in percent.", ge=0, le=100)
    gap: int | None = Field(default=None, description="Gap between packets in ms.", ge=0, le=10000)
    duplicate: float | int | None = Field(default=None, description="Probability of duplicates in percent.", ge=0, le=100)
    duplicate_corr: float | int | None = Field(default=None, description="Correlation of duplicates in percent.", ge=0, le=100)
    jitter: int | None = Field(default=None, description="Jitter of the link in ms.", ge=0, le=10000)
    reorder_prob: float | int | None = Field(default=None, description="Probability of re-orders in percent.", ge=0, le=100)
    reorder_corr: float | int | None = Field(default=None, description="Re-order correlation in percent.", ge=0, le=100)
    corrupt_prob: float | int | None = Field(
        default=None,
        description="Probability of corrupted frames in percent.",
        ge=0,
        le=100,
    )
    corrupt_corr: float | int | None = Field(default=None, description="Corruption correlation in percent.", ge=0, le=100)


class LinkConditionConfiguration(LinkCondition, extra="forbid"):
    enabled: bool = Field(default=True, description="Whether the conditioning is currently enabled.")


class LinkWithConditionConfig(LinkConnectionInfo, extra="forbid"):
    id: UUID4Type = Field(default=None, description="ID of the link.")
    label: LinkLabel = Field(default=None)
    conditioning: LinkConditionConfiguration = Field(default=None)


class LinkConditionStricted(LinkCondition, extra="forbid"):
    pass


class ConditionResponse(LinkConditionConfiguration, extra="forbid"):
    operational: LinkConditionStricted | None = Field(
        default=None,
        description="Additional operational data associated with the link conditioning.",
    )


LinkCreateBody = Annotated[
    LinkCreate,
    Body(description="The body is a JSON object that indicates the source " "and destination interfaces of the link to be created."),
]

LinkResponse = Annotated[Link, Field(description="The response body is a JSON link object.")]



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from typing import Annotated

from fastapi import Body
from pydantic import BaseModel, Field

from cml_mcp.schemas.common import (
    COLOR_EXAMPLES_STR,
    AnnotationColor,
    BorderStyle,
    UUID4Type,
)


class SmartAnnotationBase(BaseModel, extra="allow"):
    """Configuration for a smart annotation with various styling and positioning
    attributes."""

    is_on: bool = Field(default=True, description="Indicates if the smart annotation is active or not.")
    padding: int = Field(default=35, ge=1, le=200, description="Padding around the smart annotation.")
    tag: str = Field(
        default=None,
        min_length=1,
        max_length=64,
        description="A tag associated with the smart annotation.",
    )
    label: str = Field(
        default=None,
        min_length=0,
        max_length=256,
        description="A label of the smart annotation. Defaults to the tag.",
    )
    tag_offset_x: int = Field(
        default=0,
        ge=-1000,
        le=1000,
        description="Horizontal offset of the smart annotation's tag.",
    )
    tag_offset_y: int = Field(
        default=0,
        ge=-1000,
        le=1000,
        description="Vertical offset of the smart annotation's tag.",
    )
    tag_size: int = Field(
        default=14,
        ge=1,
        le=128,
        description="Font size of the smart annotation's tag text.",
    )
    group_distance: int = Field(
        default=400,
        ge=0,
        le=10000,
        description="Distance between grouped smart annotations.",
    )
    thickness: int = Field(
        default=1,
        ge=1,
        le=32,
        description="Thickness of the smart annotation’s border or line.",
    )
    border_style: BorderStyle = Field(
        default=BorderStyle.SOLID,
        description=(
            "Border style of the smart annotation - 3 values corresponding "
            'to UI values are allowed ("" - solid; "2,2" - dotted; "4,2" - dashed).'
        ),
    )
    fill_color: AnnotationColor = Field(
        default=None,
        description=f"Fill color of the smart annotation {COLOR_EXAMPLES_STR}.",
    )
    border_color: AnnotationColor = Field(
        default="#00000080",
        description=f"Border color of the smart annotation {COLOR_EXAMPLES_STR}.",
    )
    z_index: int = Field(
        default=0,
        ge=-10240,
        le=10240,
        description="Z-index of the smart annotation for stacking order.",
    )


class SmartAnnotation(SmartAnnotationBase, extra="forbid"):
    """Smart annotation element data."""

    id: UUID4Type = Field(None, description="Unique identifier for the smart annotation.")


class SmartAnnotationUpdate(SmartAnnotationBase, extra="forbid"):
    """Parameters that the smart annotation will be updated with."""

    pass


SmartAnnotationUpdateBody = Annotated[
    SmartAnnotationUpdate,
    Body(description="A JSON object with properties to update a smart annotation. " "Only supplied properties will be updated."),
]

SmartAnnotationResponse = Annotated[SmartAnnotation, Field(description="The response body is a JSON annotation object.")]



#
# This file is part of VIRL 2
# Copyright (c) 2019-2025, Cisco Systems, Inc.
# All rights reserved.
#
from enum import StrEnum
from typing import Annotated, Literal

from fastapi import Body
from pydantic import BaseModel, Field
from pydantic_strict_partial import create_partial_model

from cml_mcp.schemas.common import (
    COLOR_EXAMPLES_STR,
    AnnotationColor,
    BorderStyle,
    UUID4Type,
)

CoordinateFloat = Annotated[float, Field(description="A coordinate (floating point).", ge=-15000, le=15000)]


class AnnotationTypeEnum(StrEnum):
    text = "text"
    rectangle = "rectangle"
    ellipse = "ellipse"
    line = "line"


class LineStyle(StrEnum):
    ARROW = "arrow"
    SQUARE = "square"
    CIRCLE = "circle"


class AnnotationBase(BaseModel):
    type: AnnotationTypeEnum = Field(..., description="Annotation element type.")
    border_color: AnnotationColor = Field(..., description=f"Border color, of the annotation {COLOR_EXAMPLES_STR}.")
    border_style: BorderStyle = Field(
        ...,
        description=(
            "String defining border style - 3 values corresponding to UI values are allowed. "
            '("" - solid; "2,2" - dotted; "4,2" - dashed)'
        ),
    )
    color: AnnotationColor = Field(..., description=f"Fill color of the annotation {COLOR_EXAMPLES_STR}.")
    thickness: int = Field(..., ge=1, le=32, description="Line thickness.")
    x1: CoordinateFloat = Field(..., description="Element anchor X coordinate.")
    y1: CoordinateFloat = Field(..., description="Element anchor Y coordinate.")
    z_index: int = Field(..., ge=-10240, le=10240, description="Element Z layer.")


class X2Y2Mixin(BaseModel):
    x2: CoordinateFloat = Field(..., description="Additional X value (width, radius, ..., type dependent).")
    y2: CoordinateFloat = Field(..., description="Additional Y value (height, radius, ..., type dependent).")


class RotationMixin(BaseModel):
    rotation: int = Field(..., ge=0, le=360, description="Rotation of object, in degrees.")


class TextAnnotation(AnnotationBase, RotationMixin, extra="forbid"):
    type: Literal[AnnotationTypeEnum.text]
    text_bold: bool = Field(..., description="Text style bold.")
    text_content: str = Field(..., min_length=0, max_length=8192, description="Text element content.")
    text_font: str = Field(..., min_length=0, max_length=128, description="Text element font name.")
    text_italic: bool = Field(..., description="Text style italic.")
    text_size: int = Field(..., ge=1, le=128, description="Text size in the unit specified in `text_unit`.")
    text_unit: Literal["pt", "px", "em"] = Field(..., description="Unit of the given text size.")


class RectangleAnnotation(AnnotationBase, RotationMixin, X2Y2Mixin, extra="forbid"):
    type: Literal[AnnotationTypeEnum.rectangle]
    border_radius: int = Field(..., ge=0, le=128, description="Border radius for rectangles")


class EllipseAnnotation(AnnotationBase, RotationMixin, X2Y2Mixin, extra="forbid"):
    type: Literal[AnnotationTypeEnum.ellipse]


class LineAnnotation(AnnotationBase, X2Y2Mixin, extra="forbid"):
    type: Literal[AnnotationTypeEnum.line]
    line_start: LineStyle | None = Field(..., description="Line arrow start style.")
    line_end: LineStyle | None = Field(..., description="Line arrow end style.")


AnnotationCreate = Annotated[
    TextAnnotation | RectangleAnnotation | EllipseAnnotation | LineAnnotation,
    Field(discriminator="type"),
]

TextAnnotationUpdate = create_partial_model(model=TextAnnotation, required_fields=["type"])
RectangleAnnotationUpdate = create_partial_model(model=RectangleAnnotation, required_fields=["type"])
EllipseAnnotationUpdate = create_partial_model(model=EllipseAnnotation, required_fields=["type"])
LineAnnotationUpdate = create_partial_model(model=LineAnnotation, required_fields=["type"])

AnnotationUpdate = Annotated[
    TextAnnotationUpdate | RectangleAnnotationUpdate | EllipseAnnotationUpdate | LineAnnotationUpdate,
    Field(discriminator="type"),
]

AnnotationCreateBody = Annotated[
    TextAnnotation | RectangleAnnotation | EllipseAnnotation | LineAnnotation,
    Body(discriminator="type"),
]

AnnotationUpdateBody = Annotated[
    TextAnnotationUpdate | RectangleAnnotationUpdate | EllipseAnnotationUpdate | LineAnnotationUpdate,
    Body(discriminator="type"),
]

AnnotationUuidDescription = "Annotation Unique identifier."


class TextAnnotationResponse(TextAnnotation, extra="forbid"):
    id: UUID4Type = Field(..., description=AnnotationUuidDescription)


class RectangleAnnotationResponse(RectangleAnnotation, extra="forbid"):
    id: UUID4Type = Field(..., description=AnnotationUuidDescription)


class EllipseAnnotationResponse(EllipseAnnotation, extra="forbid"):
    id: UUID4Type = Field(..., description=AnnotationUuidDescription)


class LineAnnotationResponse(LineAnnotation, extra="forbid"):
    id: UUID4Type = Field(..., description=AnnotationUuidDescription)


AnnotationResponse = Annotated[
    TextAnnotationResponse | RectangleAnnotationResponse | EllipseAnnotationResponse | LineAnnotationResponse,
    Field(
        description="The response body is a JSON annotation object.",
        discriminator="type",
    ),
]
